
import math
import numpy as np
import thread
import threading
import random
import sys
import wx
import wx.richtext as rt
from wx.lib.pubsub import pub

from wx import glcanvas
import OpenGL
from OpenGL.GL import *
from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.vertex_shader import *
from OpenGL.GL.ARB.vertex_array_object import *
from OpenGL.GL.ARB.draw_instanced import *
from OpenGL.GL.ARB.instanced_arrays import *
from OpenGL.GL import shaders

import MatrixUtil as matrix

from PyTAS3DAssembly import *

from ctypes import sizeof, c_float, c_void_p, c_uint

from PyTAS3DUtil import ScreenQuad, Compass, Partition, Buffer, GraphicsWrapper

RAND_SEED = -1

PYTAS_FULLDISPLAY = 0
PYTAS_NODISPLAY   = 1

EMPTY_LOC         = 0
FRONTIER_LOC      = 1
FRONTIER_LOC_OFF  = 2

#assertMode = wx.PYAPP_ASSERT_DIALOG

# IDs
PYTAS_ID_STOP          = 100
PYTAS_ID_FOR           = 101
PYTAS_ID_FF            = 102
PYTAS_ID_ZOOMIN        = 103
PYTAS_ID_ZOOMOUT       = 104
PYTAS_ID_ZOOM_MODE     = 105
PYTAS_ID_SELECT_MODE   = 106
PYTAS_ID_DELETE        = 107
PYTAS_ID_UNDO          = 108
PYTAS_ID_STEPCOUNT     = 109

PYTAS_MODE_ZOOM   = 0
PYTAS_MODE_SELECT = 1

PYTAS_SM_STEP = 0
PYTAS_SM_FULL = 1

#----------------------------------------------------------------------------

class PyTAS3DCanvas(glcanvas.GLCanvas):
    def __init__(self, parent, idIn, frame, size, app):
        attribList = [glcanvas.WX_GL_RGBA,
                      glcanvas.WX_GL_DOUBLEBUFFER,
                      glcanvas.WX_GL_DEPTH_SIZE, 24]
        glcanvas.GLCanvas.__init__(self, parent, idIn, size=size, style=wx.SUNKEN_BORDER, attribList=attribList)

        self.frame              = frame
        self.app                = app
        self.context            = glcanvas.GLContext(self)
        
        self.size               = self.GetClientSize()
        self.aspectRatio        = 1.0
        self.backgroundColor    = [0.3, 0.3, 0.5, 1.0]
        self.graphics           = GraphicsWrapper(self.size)

        # Mouse state variables
        self.mousePosition      = None
        self.leftMouse          = False
        self.rightMouse         = False
        
        # Assembly state variables
        self.assembly           = PyTAS3DAssembly()
        self.keepGoing          = False
        self.stepMode           = PYTAS_SM_STEP
        self.displayPatterns    = []
        self.breakpoints        = []
        
        self.cameraUp         = np.array([0.0, 1.0, 0.0])
        self.cameraPosition   = np.array([0.0, 0.0, 8.0])

        self.initialized = False
        self.shiftDown = False

        self.goThread = None
        self._lock = threading.Lock()
        self.SetRandomSeed(RAND_SEED)
        
        self.SubscribeEventHandlers()
        
        self.numTilesPerStep    = 1
        self.lastNumTilesUpdate = 0
        self.drawUpdates        = PYTAS_FULLDISPLAY
        self.savedUpdates       = []

        #self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_SIZE,         self.OnSize)
        self.Bind(wx.EVT_MOVE,         self.OnMove)
        self.Bind(wx.EVT_PAINT,        self.OnPaint)
        self.Bind(wx.EVT_CHAR,         self.OnChar)
        self.Bind(wx.EVT_KEY_DOWN,     self.OnKeyDown)
        self.Bind(wx.EVT_KEY_UP,       self.OnKeyUp)
        
    def SubscribeEventHandlers(self):
        pub.getDefaultPublisher().subscribe(self.UpdateStatusbar, "updateStatusbar")

    def UnsubscribeEventHandlers(self):
        pub.getDefaultPublisher().unsubscribe(self.UpdateStatusbar, "updateStatusbar")

    def SendStatusBarUpdate(self, field, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "updateStatusbar", msg = (field, msg))

    def UpdateStatusbar(self, msg):
        self.frame.SetStatusText(msg[1], msg[0])

    def CreateNewAssembly(self):
        self.assembly = PyTAS3DAssembly()
        self.Reset(bForce=True)

    def SetStepMode(self, mode):
        self.stepMode = mode
        if mode == PYTAS_SM_FULL:
            self.app.SetNumTilesPerStepEnabled(False)
        else:
            self.app.SetNumTilesPerStepEnabled(True)

    def Reset(self, bFromFiles=True, bForce=False):
        if not bForce == True:
            dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
            if not dlg.ShowModal() == wx.ID_OK:
                return
            result = int(dlg.GetValue())
            dlg.Destroy()
            self.SetRandomSeed(result)
    
            self.app.OutputMessage("Resetting tile assembly system...")

        # make sure fast forward thread is stopped
        self.DoQuit()

        if bFromFiles == True:
            if not self.assembly.systemDefFile is None:
                self.assembly.ClearSpace()
                self.assembly.LoadTileSystem(self.assembly.systemDefFile)
        else:
            self.assembly.ClearSpace()
            self.assembly.PlaceSeed()
        
        for partition in self.graphics.partitions.itervalues():
            partition.deletePartition()
        del self.graphics.partitions
        self.graphics.partitions = {}

        for tileBuffer in self.graphics.tileBuffers:
            tileBuffer.deleteBuffer()
        del self.graphics.tileBuffers
        self.graphics.tileBuffers = []

        self.RedrawAssembly()
        cacheHitPercent = 0
        if self.assembly.cacheSearches > 0:
            cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))

        if not bForce == True:
            dlg = wx.MessageDialog(self, "Do you want to reset the camera position to the origin?", caption="Reset Camera?", style=wx.YES_NO)
            if dlg.ShowModal() == wx.ID_YES:
                self.CameraLookAt()
            dlg.Destroy()
        
        self.app.OutputMessage("Tile assembly system reset")

    def DoGo(self):
        self.app.SetNumTilesPerStepEnabled(False)
        if self.keepGoing:
            PrintDebug("ERROR: Cannot create another thread, there is already one running!", ERROR_MSG)
            return
        self.keepGoing = True
        try:
            self.goThread = threading.Thread(target=self.GoForward) # thread.start_new_thread(self.GoForward, ())
            self.goThread.start()
        except:
            PrintDebug("ERROR: Failed to create thread!", ERROR_MSG)

    def GoForward(self):
        while self.keepGoing:
            if not self.DoForwardStep():
                self.keepGoing = False
                cacheHitPercent = 0
                if self.assembly.cacheSearches > 0:
                    cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
        self.app.SetNumTilesPerStepEnabled(True)
        PrintDebug("Stopping the thread!", MSG_MSG)

    def DoForwardStep(self):
        if self.stepMode == PYTAS_SM_STEP:
            steps = self.numTilesPerStep
            if steps < 1:
                steps = 1
            for i in range(steps):
                newLoc = self.assembly.StepForward()
                if not newLoc is None:
                    if self.drawUpdates == PYTAS_FULLDISPLAY:
                        self.AddToBuffer(newLoc[0])
                        # self.AddToPartition(newLoc[0])
                    elif self.drawUpdates == PYTAS_NODISPLAY:
                        self._lock.acquire()
                        self.savedUpdates.append(newLoc[0])
                        self._lock.release()
                    if (self.keepGoing == False) or ((self.assembly.numTiles - self.lastNumTilesUpdate) >= 1000):
                        self.lastNumTilesUpdate = self.assembly.numTiles
                    #if ((self.assembly.numTiles % 1000) >= 0) and ((self.assembly.numTiles % 1000) < steps):

                    #elif ( ((self.assembly.numTiles - lastNumTiles) % 1000) == 0):
                        #lastNumTiles = self.assembly.numTiles
                        cacheHitPercent = 0
                        if self.assembly.cacheSearches > 0:
                            cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
                    if newLoc[1] == False:
                        self.Refresh()
                        return False
                    self.CheckForBreakpointHit(newLoc[0])
                else:
                    self.Refresh()
                    return False
            if self.keepGoing == False:
                cacheHitPercent = 0
                if self.assembly.cacheSearches > 0:
                    cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
            if self.drawUpdates == PYTAS_FULLDISPLAY:
                self.Refresh()
            return True
        elif self.stepMode == PYTAS_SM_FULL:
            newLocs = self.assembly.StepForward(True)
            if (self.keepGoing == False) or ((self.assembly.numTiles - self.lastNumTilesUpdate) >= 1000):
                self.lastNumTilesUpdate = self.assembly.numTiles
                cacheHitPercent = 0
                if self.assembly.cacheSearches > 0:
                    cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
            if not newLocs is None:
                for newLoc in newLocs[0][0]:
                    if self.drawUpdates == PYTAS_FULLDISPLAY:
                        self.AddToBuffer(newLoc)
                        # self.AddToPartition(newLoc)
                    elif self.drawUpdates == PYTAS_NODISPLAY:
                        self._lock.acquire()
                        self.savedUpdates.append(newLoc)
                        self._lock.release()
                self.CheckForBreakpointHit(newLoc)
                if self.keepGoing == False:
                    cacheHitPercent = 0
                    if self.assembly.cacheSearches > 0:
                        cacheHitPercent = float(self.assembly.cacheHits/self.assembly.cacheSearches) * 100
                    self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
                self.Refresh()
                return newLocs[1]
            return False

    def CheckForBreakpointHit(self, newTile):
        if len(self.breakpoints) > 0:
            for bp in self.breakpoints:
                if bp.type == SimUtils.SIMU_BP_TYPE_TILENUM:
                    if self.assembly.numTiles == bp.tileNum:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for " + str(self.assembly.numTiles) + " tiles in the assembly.")
                elif bp.type == SimUtils.SIMU_BP_TYPE_TILETYPE:
                    if bp.tileType == newTile[1]:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for tile of type " + bp.tileType + " placed in the assembly.")
                else:
                    newCoords = newTile[0]
                    bHit = False
                    bKeepChecking = True
                    if (bp.tileLoc[0] is not None):
                        if (bp.tileLoc[0] == newCoords[0]):
                            bHit = True
                        else:
                            bKeepChecking = False
                    if (bKeepChecking == True) and (bp.tileLoc[1] is not None):
                        if (bp.tileLoc[1] == newCoords[1]):
                            bHit = True
                        else:
                            bHit = False
                            bKeepChecking = False
                    if (bKeepChecking == True) and (bp.tileLoc[2] is not None):
                        if (bp.tileLoc[2] == newCoords[2]):
                            bHit = True
                        else:
                            bHit = False
                    if bHit == True:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for tile placed in location " + str(newCoords) + " which matches " + str(bp.tileLoc))

    def CameraLookAt(self, eye = [0.0, 0.0, 20.0], center = [0.0, 0.0, 0.0], up = [0.0, 1.0, 0.0]):
        self.cameraPosition = np.array(eye)
        self.cameraCenter   = np.array(center)
        self.cameraUp       = np.array(up)

        self.graphics.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)

    def OnPaint(self, event):
        if self.drawUpdates == PYTAS_FULLDISPLAY:
            self.savedUpdates.reverse()
            while len(self.savedUpdates) > 0:
                self.AddToBuffer(self.savedUpdates.pop())
                # self.AddToPartition(self.savedUpdates.pop())

        self._lock.acquire()
        
        dc = wx.PaintDC(self)
        self.SetCurrent(self.context)
        
        if not self.initialized:
            self.graphics.InitGraphics()
            self.CameraLookAt()
            self.DoSetViewport()
            self.initialized = True

        self.graphics.DrawScene(self.backgroundColor, self.cameraPosition)

        self.SwapBuffers()
        self._lock.release()

    def AddToBuffer(self, (coords, tileName)):
        if self.NameMatchesPatterns(tileName):
            self._lock.acquire()

            if self.graphics.tileBuffers[-1].isFull():
                self.graphics.tileBuffers.append(Buffer())
                self.graphics.newBufferCount += 1
            self.graphics.tileBuffers[-1].AddTile(self.assembly.GetLoc(coords))

            self._lock.release()

    def AddToPartition(self, (coords, tileName)):
        if self.NameMatchesPatterns(tileName):
            self._lock.acquire()

            offset = tuple([x // Partition.SIZE for x in coords])

            if offset not in self.graphics.partitions:
                self.graphics.partitions[offset] = Partition(offset)
                self.graphics.newPartitionList.append(offset)
            self.graphics.partitions[offset].AddTile(self.assembly.GetLoc(coords))

            self._lock.release()

    def RedrawAssembly(self, checkAll=False):
        self._lock.acquire()
        self.drawnAssembly = []

        for tile in self.assembly.grid.itervalues():
            if (not tile == FRONTIER_LOC) and (not tile == FRONTIER_LOC_OFF) and self.NameMatchesPatterns(tile.type.name):
                # offset = tuple([x // Partition.SIZE for x in tile.coords])
                # if offset not in self.graphics.partitions:
                #     self.graphics.partitions[offset] = Partition(offset)
                #     self.graphics.partitions[offset].initGraphics()
                # self.graphics.partitions[offset].AddTile(tile)

                if len(self.graphics.tileBuffers) == 0 or self.graphics.tileBuffers[-1].isFull():
                    self.graphics.tileBuffers.append(Buffer())
                    self.graphics.newBufferCount += 1
                self.graphics.tileBuffers[-1].AddTile(tile)

        self._lock.release()
        self.Refresh()

    def DoSetViewport(self):
        size = self.size = self.GetClientSize()
        self.graphics.size = size
        self.SetCurrent(self.context)
        glViewport(0, 0, size.width, size.height)
        self.aspectRatio = size.width * 1.0 / size.height
        self.graphics.projectionMatrix = matrix.perspective_matrix(60.0, self.aspectRatio, 0.5, 10000.0)

    def NameMatchesPatterns(self, name):
        if len(self.displayPatterns) == 0:
            return True
        else:
            for pattern in self.displayPatterns:
                if pattern in name:
                    return True
            return False

    def DoQuit(self):
        self.keepGoing = False

    def SetRandomSeed(self, seed):
        self.randSeed = seed
        if self.randSeed < 0:
            self.randSeed = random.randint(0, sys.maxint)
        self.app.OutputMessage("Seeding random number generator with " + str(self.randSeed))
        random.seed(self.randSeed)

    def ScrollToCoords(self, coords):
        #This is not (yet?) implemented
        return

    def DoSave(self):
        openFileDialog = wx.FileDialog(self.app.window, "Save", "", "", "Tile system files (*.tdp)|*.tdp",
                                        wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if len(fileName) > 0:
            assemblyString = self.assembly.AssemblyToString()
            file = open(fileName, 'w')
            file.write(assemblyString)
            file.close()

    def OnSize(self, event):
        wx.CallAfter(self.DoSetViewport)
        wx.CallAfter(self.graphics.ResizeFramebuffer)
        wx.CallAfter(self.Refresh)

    def OnMove(self, event):
        wx.CallAfter(self.DoSetViewport)
        wx.CallAfter(self.Refresh)

    def SetBackgroundColor(self):
        colorDialog = wx.ColourDialog(self.frame)
        ok = colorDialog.ShowModal()
        color = colorDialog.GetColourData().GetColour()

        self.backgroundColor = [color.Red() / 255.0, color.Green() / 255.0, color.Blue() / 255.0, 1.0]
        self.Refresh()

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        #elif keycode == ord("p") or keycode == ord("P"):
            #self.PrintGrowth()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset(False)
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSave()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad()
        elif keycode == ord("=") or keycode == ord("+"):
            self.ZoomThenRecenter(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.ZoomThenRecenter(False)
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("q") or keycode == ord("Q"):
            self.DoQuit()
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("d") or keycode == ord("D"):
            self.ToggleDisplay()
        elif keycode == wx.WXK_F2:
            self.Refresh()
        elif keycode == wx.WXK_RETURN:
            self.DeleteSelection()

        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def OnKeyDown(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_SHIFT:
            self.shiftDown = True

    def OnKeyUp(self, event):
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_SHIFT:
            self.shiftDown = False

    def OnMouse(self, event):
        # coordinates of current mouse event
        newPosition = [event.GetX(), event.GetY()]
        scroll = event.GetWheelRotation()

        glBindFramebuffer(GL_FRAMEBUFFER, self.graphics.framebuffer)
        glReadBuffer(GL_COLOR_ATTACHMENT1)
        data = glReadPixels(newPosition[0], self.size.height - newPosition[1], 1, 1, GL_RGBA_INTEGER, GL_INT)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
        
        if data[0][0][3] >= 0:
            coords = (data[0][0][0], data[0][0][1], data[0][0][2])
            loc = self.assembly.GetLoc(coords)
            if loc is not None:
                ttype = loc.type
                self.SendStatusBarUpdate(0, "Coords: " + str(loc.coords))
                msg = "Tile type: " + ttype.ToString(bFull=True)
            else:
                self.SendStatusBarUpdate(0, "")
                msg = ""
            self.SendStatusBarUpdate(2, msg)
        else:
            self.SendStatusBarUpdate(0, "")
            self.SendStatusBarUpdate(2, "")

        if event.LeftDown():
            self.leftMouse = True
        elif event.LeftUp():
            self.leftMouse = False
        elif event.RightDown():
            self.rightMouse = True
        elif event.RightUp():
            self.rightMouse = False
        elif event.Dragging():
            if self.leftMouse:
                if self.rightMouse or self.shiftDown:
                    self.cameraPosition -= self.cameraCenter
                    radius = np.linalg.norm(self.cameraPosition)
                    movement = np.append(np.array(np.subtract(newPosition, self.mousePosition) * 4.0 / [self.size.width, -self.size.height]), 0.0)
                    offset = -matrix.sphere_translation(self.cameraPosition, self.cameraUp, movement)
                    self.cameraPosition += self.cameraCenter
                    self.cameraCenter += offset * radius * 0.5
                    self.cameraPosition += offset * radius * 0.5
                    self.graphics.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)
                    self.Refresh()
                else:
                    self.cameraPosition -= self.cameraCenter
                    M = matrix.sphere_rotation_matrix(self.cameraPosition, self.cameraUp, self.mousePosition, newPosition, self.size.width, self.size.height)
                    self.cameraPosition = np.matmul(M, self.cameraPosition)
                    self.cameraPosition += self.cameraCenter
                    self.cameraUp = matrix.normalize(np.matmul(M, self.cameraUp))
                    self.graphics.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)
                    self.Refresh()
                
        elif not scroll == 0:
            if scroll < 0:
                self.cameraPosition -= self.cameraCenter
                self.cameraPosition /= 0.95
                self.cameraPosition += self.cameraCenter
            else:
                self.cameraPosition -= self.cameraCenter
                self.cameraPosition /= 1.05
                self.cameraPosition += self.cameraCenter
            self.graphics.viewMatrix = matrix.look_at_matrix(self.cameraPosition, self.cameraCenter, self.cameraUp)
            self.Refresh()
        # set coordinates of previous mouse event
        self.mousePosition = newPosition
